package OPENSSL;

$OPENSSL::VERSION = "1.0.5";

use strict;
use X500::DN;
use Digest::MD5;
use Socket;
use IO::Handle;
use IO::Select;
use File::Temp qw(tempfile);

require DynaLoader;
use vars ('@ISA');

@ISA = qw( DynaLoader );

bootstrap OPENSSL;

our $errmsg;
our $erropenssl;

my %param = (
             bin    => undef,
             conf   => undef,
             tmp    => undef,
             base   => undef,
             errmsg => undef
            );

## create an instance
sub new {
	my $this  = shift;
	my $class = ref($this) || $this;

	my $self = {
                %param,
        };

	bless $self, $class;

	my $data = { @_ };

	$self->setParam( @_ );
    
	if( not $self->{bin} ) {
		$self->{bin} = "/usr/bin/openssl";
	};

	if( not $self->{tmp} ) {
		$self->{tmp} = '/tmp';
	};
    
	if( not -e "$self->{bin}" ) {
		return;
	};
    
	$self->setError("");

    return $self;
}


sub setParam {
	my $this = shift;
	my $data = {@_};
	my $key;

	foreach $key ( keys %{$data} ) {

		$this->{conf}  = $data->{$key} if ( $key =~ /CONFIG/ );
		$this->{bin}   = $data->{$key} if ( $key =~ /BINARY/  );
		$this->{tmp}   = $data->{$key} if ( $key =~ /TMPDIR/ );
		$this->{DEBUG} = $data->{$key} if ( $key =~ /DEBUG/ );
		open STDERR, $data->{$key} if ( $key =~ /STDERR/ );
		$this->{$key}  = $data->{$key};
	}

	return 1;
}

sub errmsg {
    my $this = shift;
    return $errmsg;
}

sub erropenssl {
    my $this = shift;
    return $erropenssl;
}

sub setError {
	my $this    = shift;
    my $message = shift;
    my $err_openssl = shift || "";

    $errmsg = $message;
	$erropenssl = $err_openssl;

	if ($this->{DEBUG}) {
		print STDERR localtime()." OPENSSL->setError: errmsg:".$errmsg."\n";
		print STDERR localtime()." OPENSSL->setError: openssl error:".$erropenssl."\n" if($err_openssl ne "");
	}
    
	return undef;
}

sub createKey {

	my $this = shift;
	my $data = {@_};

	my $bits    = $data->{BITS};
	my $outfile = $data->{OUTFILE};
	my $algo    = $data->{ALGORITHM};
	my $passwd  = $data->{PASSWD};
	my @command = ($this->{bin}, 'genrsa');

	if ( defined $outfile && $outfile ne "" ) {
		push(@command, '-out', $outfile);
	}

	$passwd = '' unless(defined($passwd));
	if( $passwd ne '') {
	    push(@command, qw(-passout env:pass));

	    # defaults to des3 if a passwd given
	    unless(defined($algo) && $algo eq "") {
		$algo = 'des3';
	    }
	}

        # password query if given and no passwd
	if(defined($algo) && $algo ne "") {
	    push(@command, '-'.$algo);
	}

	if( defined $bits ) {
	    push(@command, $bits);
	}

	print STDERR localtime()." OPENSSL->createKey command: ",
	             join(" ", @command), "\n" if ($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [@command],
		env => {pass => $passwd},
	));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->createKey: Can't open pipe".
		                       " to OPENSSL: $!");
	}
	waitpid($pid, 0);
	my $ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( 0 != $ret) {
		return $this->setError("OPENSSL->createKey: OPENSSL failed".
		                       " (".$ret.")",$err);
	}

	unless( defined $outfile && $outfile ne "" ) {
	    return $out;
	}
	return 1;
}

sub createReq {

	my $this = shift;
	my $data = {@_};

	my $outfile = $data->{OUTFILE};
	my $outform = $data->{OUTFORM};
	my $keyfile = $data->{KEYFILE};
	my $passwd  = $data->{PASSWD};
	my @command = ($this->{bin}, qw(req -new));
	my ( $ret, $tmp, @DN );

	if( not $keyfile ) {
		return $this->setError("OPENSSL->createReq: No keyfile specified.");
	}

	if( defined $data->{DN} ) {
		@DN = @{ $data->{DN} };
	} else {
        return $this->setError("OPENSSL->createReq: No DN specified.");
    }

 	if( defined($this->{conf}) && -f $this->{conf}) {
		push(@command, '-config', $this->{conf});
	}

	push(@command, '-key', $keyfile);

	unless(defined($outform)) {
		$outform = 'PEM';
	} else {
		$outform = uc($outform);
	}
	if( $outform =~ /(PEM|DER)/) {
		push(@command, '-outform', $1);
		
		if( defined($outfile) and $outfile ne "") {
			push(@command, '-out', $outfile);
		}
	}
	elsif( $outform =~ /TXT/) {
		push(@command, qw(-noout -text));
	} else {
		return $this->setError("OPENSSL->createReq: Unknown OUTFORM");
	}

	push(@command, qw(-passin env:pass));

	print STDERR localtime()." OPENSSL->createReq: command: '",
	             join(' ', @command), "'.\n" if ($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [sub {

			$ENV{'pass'} = defined($passwd) ? $passwd : "";

			my ($pid, $fd) = pipe_open2send(cmd => [@_]);

			delete $ENV{'pass'};

			if(defined($pid)) {
				print $fd join("\n", @DN), "\n";
				close($fd);
				waitpid($pid, 0);
				my $ret = ($? >> 8);
				exit($ret);
			}
			exit(1);
		}, @command]
	));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->createReq: Can't open pipe".
		                       " to OPENSSL: $!");
	}
	waitpid($pid, 0);
	$ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( 0 != $ret) {
		return $this->setError("OPENSSL->createReq: OPENSSL failed".
		                       " (".$ret.")",$err);
	}

	unless(defined $outfile and $outfile ne "") {
		return $out; # openssl stdout data
	} else {
		return 1;    # true, outfile written
	}
}

sub createSelfSignedCert {

	my $this = shift;
	my $data = {@_};
    
	my $outfile = $data->{OUTFILE};
	my $keyfile = $data->{KEYFILE};
	my $reqfile = $data->{REQFILE};
	my $noemail = $data->{NOEMAILDN};
	my $passwd  = $data->{PASSWD};
	my $days    = $data->{DAYS};
	my @command = ($this->{bin}, qw(req -x509));

	my ( $ret, $tmp );

	unless(defined($keyfile) and $keyfile ne "") {
		return $this->setError("OPENSSL->createSelfSignedCert: No keyfile specified.");
	}
	unless(defined($reqfile) and $reqfile ne "") {
		return $this->setError("OPENSSL->createSelfSignedCert: No requestfile specified.");
	}

	## if ( defined($noemail) && $noemail ne "" ) {
	##	push(@command, '-noemailDN');
	## }

	$passwd = '' unless(defined($passwd));
	if( $passwd ne '' ) {
	    push(@command, qw(-passin env:pass));
	}

	if( defined $this->{'conf'} && $this->{conf} ne "" ) {
	    push(@command, '-config', $this->{conf});
	}

	if( defined $days && $days =~ /^\d+$/ && $days > 0) {
	    push(@command, '-days', $days);
	}

	push(@command, '-in', $reqfile, '-key', $keyfile);

	if( defined($outfile) && $outfile ne "" ) {
		push(@command, '-out', $outfile);
	}

	print STDERR localtime()." OPENSSL->createSelfSignedCert command: ",
		     join(" ", @command), "\n" if($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [@command],
		env => {pass => $passwd},
	));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->createSelfSignedCert: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	$ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( 0 != $ret) {
		return $this->setError("OPENSSL->createSelfSignedCert: ".
		                       "OPENSSL failed ".
		                       "(".$ret.")",$err);
	}

	unless( defined $outfile && $outfile ne "" ) {
	    return $out;
	}
	return 1;
}

sub issueReq {

	my $this = shift;
	my $data = {@_};

	my $reqfile  = $data->{REQFILE};
	my $cakey    = $data->{CAKEY};
	my $days     = $data->{DAYS};
	my $startDate= $data->{START_DATE};
	my $endDate  = $data->{END_DATE};
	my $passwd   = $data->{PASSWD};
	my $exts     = $data->{EXTS};

	my $outdir   = $data->{OUTDIR};
	my $outfile  = $data->{OUTFILE};
	my $caName   = $data->{CA_NAME};

	unless( defined $reqfile && $reqfile ne "" ) {
	    return $this->setError("OPENSSL->issueReq: No request specified.");
	}

	my @command = ($this->{bin}, qw(ca -batch));
	my $ret;

	if(defined($this->{conf}) and $this->{conf} ne '') {
		push(@command, '-config', $this->{conf});
	}
	if(defined($cakey) and $cakey ne '') {
		push(@command, '-keyfile', $cakey);
	}

	$passwd = '' unless(defined($passwd));
	if($passwd ne '') {
		push(@command, qw(-passin env:pass));
	}

	if( defined $days && $days =~ /^\d+$/ && $days > 0) {
	    push(@command, '-days', $days);
	}

	if( defined($exts) and $exts ne '') {
		push(@command, '-extensions', $exts);
	}

	if( defined($startDate) and $startDate ne '') {
		push(@command, '-startdate', $startDate);
	}

	if( defined($endDate) and $endDate ne '') {
		push(@command, '-startdate', $endDate);
	}

	if( defined($caName) and $caName ne '') {
		push(@command, '-startdate', $caName);
	}

	if( defined($data->{NOEMAILDN}) and "".$data->{NOEMAILDN} ne '0') {
		push(@command, '-noemailDN');
	}
	if( defined($data->{NOUNIQUEDN}) and "".$data->{NOUNIQUEDN} ne '0') {
		push(@command, '-nouniqueDN');
	}

	#this has to be the last option
	if( defined($outfile) and $outfile ne '') {
		push(@command, '-out', $outfile);
	}
	if( defined($reqfile) and $reqfile ne '') {
		push(@command, '-in', $reqfile);
	}

   	print STDERR localtime()." OPENSSL->issueReq command: ",
      join(" ", @command), "\n" if($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [@command],
		env => {pass => $passwd},
	));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->issueReq: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	$ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( 0 != $ret) {
        if(defined $outfile && $outfile ne "" && -e $outfile) {
            # issue failed, so remove broken file
            unlink($outfile);
        }
		return $this->setError("OPENSSL->issueReq: ".
		                       "OPENSSL failed ".
		                       "(".$ret.")",$err);
	}

	unless( defined $outfile && $outfile ne "" ) {
	    return $out;
	}
	return 1;
}

sub revokeCert {

	my $this = shift;
	my $data = {@_};

	my $cakey    = $data->{CAKEY};
	my $cacert   = $data->{CACERT};
	my $passwd   = $data->{PASSWD};
	my $certFile = $data->{INFILE};
	my $crlReason= $data->{CRL_REASON};

	if (!defined $certFile || $certFile eq "" ) {
		return $this->setError("OPENSSL->revokeCert: No certificate specified.");
	}

	my @command = ($this->{bin}, "ca", "-revoke", "$certFile");
    
    if( defined $this->{'conf'} && $this->{conf} ne "" ) {
        push @command , "-config", $this->{conf};
    }
    if( defined $cakey && $cakey ne "" ) {
        push @command, "-keyfile", "$cakey";
    }
    if ( defined $passwd && $passwd ne "" ) {
        push @command, "-passin", "env:pass";
    }
    if ( defined $cacert && $cacert ne "" ) {
        push @command, "-cert", $cacert;
    }
    if( $data->{NOUNIQUEDN} ) {
        push @command, "-nouniqueDN" ;
    }
    if(defined $crlReason && $crlReason ne "") {
        push @command, "-crl_reason", $crlReason;
    }

    print STDERR localtime()." OPENSSL->revokeCert command: ".join(' ',@command)."\n" if ($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
         cmd => [@command],
         env => {pass => $passwd},
    ));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->issueReq: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	my $ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( $ret != 0) {
		return $this->setError("OPENSSL->revokeCert: OPENSSL failed (".$ret.").",
                               $err);
	} else {
		return 1;
	}
}

sub issueCRL {

	my $this = shift;
	my $data = {@_};

	my $cakey    = $data->{CAKEY};
	my $cacert   = $data->{CACERT};
	my $days     = $data->{DAYS};
	my $passwd   = $data->{PASSWD};
	my $outfile  = $data->{OUTFILE};
	my $outform  = $data->{OUTFORM};
	my $exts     = $data->{EXTS};

	my @command = ($this->{bin}, "ca", "-gencrl");

	if ( defined $outfile && $outfile ne "" ) {
        push @command, "-out", $outfile;
	}
    if( defined $this->{'conf'} && $this->{conf} ne "" ) {
        push @command, "-config", $this->{conf};
    }
    if( defined $cakey && $cakey ne "" ) {
        push @command, "-keyfile", "$cakey";
    }
    if ( defined $passwd && $passwd ne "" ) {
        push @command, "-passin", "env:pass";
    }
    if ( defined $cacert && $cacert ne "" ) {
        push @command, "-cert", $cacert;
    }
    if ( defined $days && $days ne "" ) {
        push @command, "-crldays", $days;
    }
    if ( defined $exts && $exts ne "" ) {
        push @command, "-crlexts", $exts;
    }
    if ( $data->{NOUNIQUEDN} ) {
        push @command, "-nouniqueDN";
    }

    print STDERR localtime()." OPENSSL->issueCRL command: ".join(' ', @command)."\n" if($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
         cmd => [@command],
         env => {pass => $passwd},
    ));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->issueReq: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	my $ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	if( $ret != 0) {
		return $this->setError("OPENSSL->issueCRL: OPENSSL failed (".$ret.").",
                               $err);
	}

	if( !defined $outfile ) {
		return $out;
	}
	return 1;
}

sub convert {

	my $this = shift;
	my $data = {@_};

	my $indata  = $data->{DATA};
	my $type    = $data->{DATATYPE};
	my $outfile = $data->{OUTFILE};
	my $outform = $data->{OUTFORM};
	my $inform  = $data->{INFORM};
	my $infile  = $data->{INFILE};
	my $keyfile = $data->{KEYFILE};
	my $inpasswd  = $data->{'INPASSWD'};
	my $outpasswd = $data->{'OUTPASSWD'};
	my $algo    = $data->{'ALGO'} || 'des3';
	my $cacert  = $data->{'CACERT'};
    my $chain   = $data->{'CHAIN'} || undef;
    my $capath  = $data->{'CAPATH'} || undef;
	my $nokeys  = $data->{'NOKEYS'};

	my ( $tmp_fd, $tmp_file, $ret );
	my @command = ($this->{bin});

	## rest errordetection
	if( $? != 0 ) {
		print STDERR localtime()." OPENSSL->convert: resetting error from ${?} to 0\n" if ($this->{DEBUG});
		$? = 0;
	}

	if ( !defined $type || $type eq "") {
		return $this->setError("OPENSSL->convert: No datatype specified.");
	}
	if ( not $algo =~ /des3|des|idea/ ) {
		return $this->setError("OPENSSL->convert: Unsupported algorithm specified.");
	}
	if(defined $infile and $infile ne "") {
		# Return if $infile does not exists
		if( not -e $infile ) {
			return $this->setError("OPENSSL->convert: The specified inputfile doesn't exist.");
		}
		$tmp_file = undef;
	} else {
		# Return if no data buffer given
		unless(defined($indata) and length($indata)) {
			return $this->setError("OPENSSL->convert: No input data or file specified");
		}

		( $tmp_fd, $tmp_file) = tempfile();
		unless($tmp_fd and $tmp_file) {
			return $this->setError("OPENSSL->convert: Can't open temporary file.");
		}
		if( !defined(write_data($tmp_fd, $indata))) {
			close($tmp_fd);
			unlink($tmp_file);
			return $this->setError("OPENSSL->convert: Can't write temporary file.");
		}
		close($tmp_fd);
		$infile = $tmp_file;
	}

	$outform = "PEM" if( not $outform ); 
	$inform  = "PEM" if( not $inform ); 

    # TYPE
    if( $type =~ /CRL/i ) {
		push(@command, 'crl');
    } elsif ( $type =~ /CERTIFICATE/i ) {
		if( $outform eq 'PKCS12' || $inform eq 'PKCS12' ) {
		    push(@command, 'pkcs12');
		} else {
		    push(@command, 'x509', '-nameopt', 'RFC2253');
		}
    } elsif ( $type =~ /REQ/i ) {
		push(@command, 'req', '-nameopt', 'RFC2253');
        if (defined $this->{conf} && $this->{conf} ne "") { 
            push(@command, '-config', $this->{conf});
        }
	} elsif ( $type =~ /KEY/i ) {
		push(@command, 'rsa');
        if( defined($inpasswd)) {
			push(@command, '-passin', 'env:inpass');
        }
        if( defined $outpasswd && $outpasswd ne "") {
			push(@command, '-passout', 'env:outpass');
        }
    } else {
		# unknown type...
		$this->setError("OPENSSL->convert: The datatype which should be converted is not known.");
		return undef;
	}

	# FILES
	if(defined $infile  and $infile ne "") {
		push(@command, '-in',  $infile);
	}
	if(defined $outfile and $outfile ne "") {
		push(@command, '-out', $outfile);
	}

    # INFORM
	if( $outform ne 'PKCS12' ) {
        if( $inform =~ /(PEM|DER|NET)/i ) {
			push(@command, '-inform', uc($inform));
        } elsif( $inform eq 'PKCS12' ) {
			if( defined($nokeys) ) {
				push(@command, '-nokeys');
			}
			# passin is the p12pass
			if( defined($inpasswd) ) {
				push(@command, '-passin', 'env:inpass');
			}
			# passout is for keys we extract
			if( defined($outpasswd) and $outpasswd ne "") {
				push(@command, '-passout', 'env:outpass');
				push(@command, '-'.$algo)
					if( $algo eq 'des'  ||
						$algo eq 'des3' ||
						$algo eq 'idea' );
            } else {
				push(@command, '-nodes');
			}
        } else {
            ## unknown inform
            unlink($tmp_file) if($tmp_file);
            return $this->setError("OPENSSL->convert: input format is unknown or unsupported.");
        }
    }

    # OUTFORM
	if( $outform =~ /TXT/i ) {
		push(@command, '-text', '-noout');
	} elsif ( $outform =~ /(PEM|DER|NET)/i ) {
		if($inform ne 'PKCS12') {
			# PKCS12 outform is _PEM_
			push(@command, '-outform', uc($outform));
		}
    } elsif ( $outform eq 'PKCS12' ) {
		push(@command, '-export');
        if(defined $chain && defined $capath && $capath ne "") {
            push(@command, '-chain');
            push(@command, '-CApath', $capath);
        }
		# passout is for p12 file
		if(defined($outpasswd) and $outpasswd ne "") {
			push(@command, '-passout', 'env:outpass');
		}
		# passin is for key files
		if(defined($inpasswd) ) {
			push(@command, '-passin',  'env:inpass');
		}
		push(@command, '-inkey',    $keyfile) if(defined $keyfile);
		push(@command, '-certfile', $cacert)  if(defined($cacert));
	} else {
		## unknown outform
		unlink($tmp_file) if($tmp_file);
		return $this->setError("OPENSSL->convert: The output format is unknown or unsupported.");
	}

	print STDERR localtime()." OPENSSL->convert: command: '",
	             join(' ', @command), "'.\n"; # if ($this->{DEBUG});

	my ($pid, $out, $err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [sub {
			my @cmd = @_;
			my %env = (
				inpass => defined($inpasswd)  ? $inpasswd  : "",
				outpass=> defined($outpasswd) ? $outpasswd : "",
			);

			exec_wrapper(\@cmd, \%env);
			exit(1);
		}, @command]
	));
	unless(defined($pid)) {
		unlink($tmp_file) if($tmp_file);
		return $this->setError("OPENSSL->convert: Can't open pipe".
		                       " to OPENSSL: $!");
	}
	waitpid($pid, 0);
	$ret = ($? >> 8);
	$err = '' unless(defined($err));
	#$out = '' unless(defined($out));

	unlink($tmp_file) if($tmp_file);
	if(defined $outfile and $outfile ne "") {
		unless( $ret == 0 || ($ret == 1 && $err eq "")) {
			unlink($outfile); # remove broken file
			return $this->setError("OPENSSL->convert: OPENSSL failed".
			                       " (".$ret.")",$err);
		}
		return 1;    # true, outfile written
	} else {
		unless( $ret == 0 || ($err eq "" && $out ne "")) {
			return $this->setError("OPENSSL->convert: OPENSSL failed".
			                       " (".$ret.")",$err);
		}
		return $out; # openssl stdout data
	}
}

sub getCertAttribute {
	my $this = shift;
	my $data = {@_};
    
	my $cert;
	my @attribute = ();
	my $ret;

	if( $data->{ATTRIBUTE_LIST} && ref($data->{ATTRIBUTE_LIST}) ) {
		@attribute = @{$data->{ATTRIBUTE_LIST}};
	} else {
        return this->setError("OPENSSL->getCertAttribute: no ATTRIBUTE_LIST.");
	}

	if($data->{INFORM} && $data->{INFORM} =~ /DER/)
	{
		$cert = OPENSSL::X509::_from_der($data->{DATA});
	} else {
		$cert = OPENSSL::X509::_from_pem($data->{DATA});
	}

	if (not $cert) {
        return $this->setError("OPENSSL->getCertAttribute: can not convert data");
    }

	foreach my $attribute ( @attribute ) {
		my $function = $attribute;
        if($function eq "dn") {
            $function = "subject";
        } elsif($function eq "hash") {
            $function = "subject_hash";
        }
        $ret->{uc($attribute)} = $cert->$function;
	}
	return $ret;
}

sub getReqAttribute {
	my $this = shift;
	my $data = {@_};

	my $csr;
	my $ret;

	my @attribute = ();
	if( $data->{ATTRIBUTE_LIST} && ref($data->{ATTRIBUTE_LIST}) ) {
		@attribute = @{$data->{ATTRIBUTE_LIST}};
	} else {
        return this->setError("OPENSSL->getReqAttribute: no ATTRIBUTE_LIST.");
	}

	if ($data->{INFORM} and $data->{INFORM} =~ /DER/)
	{
		$csr = OPENSSL::REQ::_from_der($data->{DATA});
	} else {
		$csr = OPENSSL::REQ::_from_pem($data->{DATA});
	}

	if (not $csr) {
        return $this->setError("OPENSSL->getReqAttribute: can not convert data");
    }

	foreach my $attribute ( @attribute ) {
		my $function = $attribute;
        if($function eq "dn") {
            $function = "subject"
        };
        $ret->{uc($attribute)} = $csr->$function;
	}

	return $ret;
}

sub getCRLAttribute {
	my $this = shift;
	my $data = {@_};

	my $crl;
	my $ret;

	my @attribute = ();
	if( $data->{ATTRIBUTE_LIST} && ref($data->{ATTRIBUTE_LIST}) ) {
		@attribute = @{$data->{ATTRIBUTE_LIST}};
	} else {
        return this->setError("OPENSSL->getCRLAttribute: no ATTRIBUTE_LIST.");
	}

	if ($data->{INFORM} and $data->{INFORM} =~ /DER/) {
		$crl = OPENSSL::CRL::_from_der($data->{DATA});
	} else {
		$crl = OPENSSL::CRL::_from_pem($data->{DATA});
	}

	if (not $crl) {
        return $this->setError("OPENSSL->getCRLAttribute: can not convert data");
    }

	foreach my $attribute ( @attribute ) {
		my $function = $attribute;
        if( $function eq "dn") {
            $function = "issuer";
        }
		$ret->{uc($attribute)} = $crl->$function;
	}
	return $ret;
}

sub updateDB {

	my $this = shift;
	my $data = {@_};

	my $cakey    = $data->{CAKEY};
	my $cacert   = $data->{CACERT};
	my $passwd   = $data->{PASSWD};
	my $outfile  = $data->{OUTFILE};

	my ( $ret, $tmp );
    my @command = ($this->{bin}, "ca", "-updatedb");

    if( defined $this->{'conf'} && $this->{conf} ne "" ) {
        push @command, "-config", $this->{conf};
    }
    if( defined $cakey && $cakey ne "" ) {
        push @command, "-keyfile", $cakey;
    }
    if( defined $passwd && $passwd ne "" ) {
        push @command, "-passin", "env:pass";
    }
    if( defined $cacert && $cacert ne "" ) {
        push @command, "-cert", $cacert;
    }

    print STDERR localtime()." OPENSSL->updateDB command: ".join(" ", @command)."\n" if ($this->{DEBUG});

	my ($pid, $out, $osl_err) = pipe_recv2_data(pipe_open2recv2(
		cmd => [@command],
		env => {pass => $passwd},
	));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->updateDB: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	$ret = ($? >> 8);
	$osl_err = '' unless(defined($osl_err));
    $out = '' unless(defined($out));

	if( ($ret != 0 && $ret != 1) || $osl_err =~ /error/g ) {
		return $this->setError("OPENSSL->updateDB: OPENSSL failed (".$?.").",
                               $osl_err);
	}
	return 1;
}

sub verify {
    my $this = shift;
    my $data = {@_};

    my $cert       = $data->{'CERT'};
    my $CApath     = $data->{'CAPATH'}   || undef;
    my $doCrlCheck = $data->{'CRLCHECK'} || undef;
    
    my $retValue = "";
    my $errorMsg = "";
    my $errorNmb = 0;
        
    if(not defined $cert) {
        return $this->setError("No certificate to verify!");
    }

    my @command = ($this->{bin}, "verify");

    if(defined $CApath && $CApath ne "" ) {
        push @command, "-CApath", $CApath;
    }
    if(defined $doCrlCheck) {
        push @command, "-crl_check";
    }
    push @command, $cert;
    
    print STDERR localtime()." OPENSSL->verify command: ".join(" ", @command)."\n" if ($this->{DEBUG});

	my ($pid, $out, $osl_err) = pipe_recv2_data(pipe_open2recv2(
                                                                cmd => [@command]
                                                               ));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->verify: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	my $ret = ($? >> 8);
	$osl_err = '' unless(defined($osl_err));
    $out = '' unless(defined($out));

    foreach my $line (split(/\n/, $out)) {
        
        $line =~ /^\/.*\.pem:\s+(.*)\s*$/;
        if( defined $1 && $1 eq "OK" ) {
            $retValue = "OK";
            last;
        }

        $line =~ /^error\s+(\d+)\s+at\s+\d+\s+[\w\s]+:(.*)$/;
        
        if(defined $1 && defined $2 ) {
            chomp($line);
            $retValue = $2;
            $errorMsg = $line;
            $errorNmb = $1;
        }
    }
    if($retValue eq "OK") {
        return 1;
    } else {
        if(defined $osl_err && $osl_err ne "") {
            return $this->setError("OPENSSL->verify: Certificate invalid! ($retValue / $errorMsg)",
                                   $osl_err);
        } else {
            return $this->setError("OPENSSL->verify: Certificate invalid! ($retValue / $errorMsg)");
        }
    }
}

sub status {
    my $this = shift;
    my $data = {@_};

    my $serial = $data->{SERIAL};

    if(! defined $serial || $serial !~ /^[[:xdigit:]]+$/) {
        return $this->setError("OPENSSL->status: wrong serial number.");
    }

    my @command = ($this->{bin}, "ca");
    
    if( defined $this->{'conf'} && $this->{conf} ne "" ) {
        push @command, "-config", $this->{conf};
    }
    push @command, "-status", $serial;

    print STDERR localtime()." OPENSSL->status command: ".join(" ", @command)."\n" if ($this->{DEBUG});

	my ($pid, $out, $osl_err) = pipe_recv2_data(pipe_open2recv2(
                                                                cmd => [@command]
                                                               ));
	unless(defined($pid)) {
		return $this->setError("OPENSSL->status: ".
		                       "Can't open pipe to OPENSSL: $!");
	}
	waitpid($pid, 0);
	my $ret = ($? >> 8);
	$osl_err = '' unless(defined($osl_err));
    $out = '' unless(defined($out));

    my $err = "";
    foreach my $line (split(/\n/, $osl_err)) {
        next if($line =~ /^Using configuration from/);
        next if($line =~ /^DEBUG/);
        if($line =~ /^$serial=(\w+)\s+.*$/) {
            return $1;
        } else {
            $err .= $line."\n";
        }
    }
    return $this->setError("OPENSSL->status: Show certificate status with serial '$serial' failed.($ret)",
                            $err);
}

sub getOPENSSLDate {
	my $this = shift;
    my $date = shift;

	if (not defined $date) {
		return $this->setError("OPENSSL->getOPENSSLDate: No date specified.");
	}
	my $Date = $this->getNumericDate ( $date );
	if (not defined $Date) {
		return undef;
	}

	## remove the century
	$Date =~ s/^..//;

	## add the trailing Z
	$Date .= "Z";

	return $Date; 
}

sub getNumericDate {
	my $this = shift;
    my $date = shift;

	if (not defined $date) {
		return $this->setError("OPENSSL->getNumericDate: No date specified.");
	}
	my %dummy;
	my $new_date;

	##  The Format looks like this: May 13 14:32:12 2004 GMT

	## remove the leading days like SUN or MON
	if ( $date =~ /^\s*[^\s]+\s+(JAN|FEB|MAR|APR|MAY|JUN|JUL|AUG|SEP|OCT|NOV|DEC)/i ) {
		$date =~ s/^\s*[^\s]+//;
	}

	## Month
	if ( $date =~ /^\s*JAN/i ) {
		$dummy {MONTH} = "01";
	} elsif ( $date =~ /^\s*FEB/i ) {
		$dummy {MONTH} = "02";
	} elsif ( $date =~ /^\s*MAR/i ) {
		$dummy {MONTH} = "03";
	} elsif ( $date =~ /^\s*APR/i ) {
		$dummy {MONTH} = "04";
	} elsif ( $date =~ /^\s*MAY/i ) {
		$dummy {MONTH} = "05";
	} elsif ( $date =~ /^\s*JUN/i ) {
		$dummy {MONTH} = "06";
	} elsif ( $date =~ /^\s*JUL/i ) {
		$dummy {MONTH} = "07";
	} elsif ( $date =~ /^\s*AUG/i ) {
		$dummy {MONTH} = "08";
	} elsif ( $date =~ /^\s*SEP/i ) {
		$dummy {MONTH} = "09";
	} elsif ( $date =~ /^\s*OCT/i ) {
        $dummy {MONTH} = "10";
	} elsif ( $date =~ /^\s*NOV/i ) {
		$dummy {MONTH} = "11";
	} elsif ( $date =~ /^\s*DEC/i ) {
		$dummy {MONTH} = "12";
	} else {
		return $this->setError("OPENSSL->getNumericDate: Illelgal month. ($date)");
	}

	## day
	$date =~ s/^ *//;
	$date = substr ($date, 4, length ($date)-4);
	$dummy {DAY}    = substr ($date, 0, 2);
	$dummy {DAY}    =~ s/ /0/;
	$dummy {HOUR}   = substr ($date, 3, 2);
	$dummy {MINUTE} = substr ($date, 6, 2);
	$dummy {SECOND} = substr ($date, 9, 2);
	$dummy {YEAR}   = substr ($date, 12, 4);

	## build the new date
	$new_date =	$dummy {YEAR}.$dummy {MONTH}.$dummy {DAY}.
                $dummy {HOUR}.$dummy {MINUTE}.$dummy {SECOND};

	return $new_date; 
}

sub getParsedCert {
	my $this = shift;
    my $data = {@_};

    my $Data   = $data->{'DATA'};
    my $infile = $data->{'INFILE'};
    my $inform = $data->{'INFORM'};

    my $tmp;

    $this->{certFormat} = ( $inform || "PEM" );

    if( $infile ) {
        $this->{item} = "";
		open( FD, "<$infile" )
          or return $this->setError("OPENSSL->getParsedCert: Cannot open infile '$infile' for reading.");
		while ( $tmp = <FD> ) {
			$this->{item} .= $tmp;
		}
		close( FD );
	} else {
        $this->{item} = $Data;
    }
    
	if ( defined $this->{item} && $this->{item} ne "" ) {
        my $beginCert = "-----BEGIN[\\w\\s]+[-]{5}";
        my $endCert   = "-----END[\\w\\s]+[-]{5}";
        ( $this->{cert} ) = ( $this->{'item'} =~ /($beginCert[\S\s\n]+$endCert)/ );
        
        $tmp = "";
        $tmp = $this->parseCert();
		if ( not $tmp ) {
			return $this->setError("OPENSSL->getParsedCert: Cannot initialize certificate", 
                                   $this->erropenssl());
		} else {
            return $tmp;
        }
    }
    return $this->setError("OPENSSL->getParsedCert: The certificate was not parsed.")
}

sub getTXTCert {
    my $this = shift;
    my $data = {@_};
    
    my $Data   = $data->{'DATA'}   || undef;
    my $infile = $data->{'INFILE'} || undef;
    my $inform = $data->{'INFORM'} || "PEM";

    my %hash = (
                DATATYPE => "CERTIFICATE",
                INFORM   => $inform,
                OUTFORM  => "TXT"
                );
    if(defined $infile && $infile ne "" && -e $infile) {
        $hash{'INFILE'} = $infile;
    } elsif(defined $Data && $Data ne "") {
        $hash{'DATA'} = $Data;
    } else {
        return $this->setError("No certificate found!");
    }
    
    my $ret = $this->convert( %hash );
    if(not defined $ret) {
        return $this->setError("Cannot convert certificate to TXT-format ".
                               "(".$this->errmsg().")", $this->erropenssl());
    }
    return $ret;
}

sub getPEMCert {
    my $this = shift;
    
    if ( $this->{certFormat} eq 'PEM' ) {
        $this->{cert} =~ s/^\n*//;
        $this->{cert} =~ s/\n*$/\n/;
        return $this->{cert};
    }
    if (not $this->{pemCert}) {
        $this->{pemCert} = $this->convert( DATA=>$this->{cert},
                                           DATATYPE=>"CERTIFICATE",
                                           INFORM=>$this->{certFormat},
                                           OUTFORM=>"PEM" );
        if(not $this->{pemCert}){
            return $this->setError("OPENSSL->getPEMCert: Cannot convert request to PEM-format ".
                                   "(".$OPENSSL::errmsg.")", $this->erropenssl())
        }
    }
    return $this->{pemCert};
}

sub getDERCert {
    my $this = shift;
    
    if( $this->{certFormat} eq 'DER' ) {
        return $this->{cert};
    }
    if(not $this->{derCert}) {
        $this->{derCert} = $this->convert( DATA=>$this->{cert},
                                           DATATYPE=>"CERTIFICATE",
                                           INFORM=>$this->{certFormat},
                                           OUTFORM=>"DER" );
        if(not $this->{derCert}) {
            return $this->setError("OPENSSL->getDERCert: Cannot convert request to DER-format ".
                                   "(".$OPENSSL::errmsg.")", $this->erropenssl());
        }
    }
    return $this->{derCert};
}

sub parseCert {
	my $this = shift;

    if (not $this->{cert}) {
        return $this->setError("OPENSSL->parseCert: No certificate present.");
    }
    
	my ( @ouList, @exts, $ret, $k, $v, $tmp, $md5 );
    
	my @attList = ( "serial", "dn", "issuer", "notBefore", "notAfter",
                    "alias", "modulus", "pubkey", "fingerprint", "hash", "emailaddress",
                    "version", "pubkey_algorithm", "signature_algorithm", "exponent",
                    "keysize", "extensions" );
	if ($this->{certFormat} eq "DER") {
		$ret = $this->getCertAttribute(
                                       ATTRIBUTE_LIST => \@attList,
                                       DATA           => $this->getDERCert(),
                                       INFORM         => "DER");
	} else {
		$ret = $this->getCertAttribute(
                                       ATTRIBUTE_LIST => \@attList,
                                       DATA           => $this->getPEMCert(),
                                       INFORM         => "PEM");
	}
    if(not $ret) {
        return $this->setError("OPENSSL->parseCert: Can not parse Certificate.");
    }

	$ret->{DN} =~ s/(^\/|\/$)//g;
	$ret->{DN} =~ s/\/([A-Za-z0-9\-]+)=/, $1=/g;
    
	$ret->{ISSUER} =~ s/(^\/|\/$)//g;
	$ret->{ISSUER} =~ s/\/([A-Za-z0-9\-]+)=/, $1=/g;

	my $i = 0;
	my $acton = "name";
	while($i < length($ret->{DN})) {
		if (substr($ret->{DN}, $i, 1) =~ /\\/) {
			$i++;
		} elsif(substr($ret->{DN}, $i, 1) =~ /=/) {
			if($acton eq "value") {
				$ret->{DN} = substr($ret->{DN}, 0, $i)."\\".substr($ret->{DN}, $i);
				$i++;
			} else {
				$acton = "value";
			}
		} elsif(substr($ret->{DN}, $i, 1) =~ /[,+]/) {
			$acton = "name";
		}
		$i++;
	}

	## split DN into DN_HASH
	if ($ret->{DN} =~ /\\/) {
		my $x500_dn = X500::DN->ParseRFC2253($ret->{DN});
		if(not $x500_dn) {
			return $this->setError("OPENSSL->parseCert: X500::DN failed.");
		}
		my $rdn;
		foreach $rdn($x500_dn->getRDNs()) {
			next if($rdn->isMultivalued());
			my @attr_types = $rdn->getAttributeTypes();
			my $type  = $attr_types[0];
			my $value = $rdn->getAttributeValue($type);
			push (@{$ret->{DN_HASH}->{uc($type)}}, $value);
		}
	} else {
		my @rdns = split(/,/, $ret->{DN});
		foreach my $rdn (@rdns) {
			my ($type, $value) = split(/=/, $rdn);
			$type =~ s/^\s*//;
			$type =~ s/\s*$//;
			$value =~ s/^\s*//;
			$value =~ s/\s*$//;
			push (@{$ret->{DN_HASH}->{uc($type)}}, $value);
		}
	}
    
    # Do the same for the issuer

	$i = 0;
	$acton = "name";
	while ($i < length($ret->{ISSUER})) {
		if (substr($ret->{ISSUER}, $i, 1) =~ /\\/) {
			$i++;
		} elsif(substr($ret->{ISSUER}, $i, 1) =~ /=/) {
			if ($acton eq "value") {
				$ret->{ISSUER} = substr($ret->{ISSUER}, 0, $i)."\\".substr($ret->{ISSUER}, $i);
				$i++;
			} else {
				$acton = "value";
			}
		} elsif(substr ($ret->{ISSUER}, $i, 1) =~ /[,+]/) {
			$acton = "name";
		}
		$i++;
	}
    
	## split ISSUER into ISSUER_HASH
	if ($ret->{ISSUER} =~ /\\/) {
		my $x500_dn = X500::DN->ParseRFC2253($ret->{ISSUER});
		if (not $x500_dn) {
			return $this->setError("OPENSSL->parseCert: X500::DN failed.");
		}
		my $rdn;
		foreach $rdn ($x500_dn->getRDNs()) {
			next if($rdn->isMultivalued());
			my @attr_types = $rdn->getAttributeTypes();
			my $type  = $attr_types[0];
			my $value = $rdn->getAttributeValue($type);
			push (@{$ret->{ISSUER_HASH}->{uc($type)}}, $value);
		}
	} else {
		my @rdns = split(/,/, $ret->{ISSUER});
		foreach my $rdn (@rdns) {
			my ($type, $value) = split(/=/, $rdn);
			$type =~ s/^\s*//;
			$type =~ s/\s*$//;
			$value =~ s/^\s*//;
			$value =~ s/\s*$//;
			push (@{$ret->{ISSUER_HASH}->{uc($type)}}, $value);
		}
	}
    
	if( exists $ret->{PUBKEY} ) {
		$md5 = new Digest::MD5;
		$md5->add( $ret->{PUBKEY} );
		$ret->{KEY_DIGEST} = $md5->hexdigest();
	}

	## Check if Email field is only present in subjectAltName
	if (! $ret->{EMAILADDRESS} &&
	    exists $ret->{DN_HASH}->{EMAILADDRESS} &&
	    $ret->{DN_HASH}->{EMAILADDRESS}[0]) {
		$ret->{EMAILADDRESS} = $ret->{DN_HASH}->{EMAILADDRESS}[0];
	}

	## load all extensions
	$ret->{PLAIN_EXTENSIONS} = $ret->{EXTENSIONS};
	delete $ret->{EXTENSIONS};
	$ret->{OPENSSL_EXTENSIONS} = {};

	my ($c, $val, $key);
	my @lines = split(/\n/, $ret->{PLAIN_EXTENSIONS});
    
	$i = 0;
	while($i < @lines) {
		if($lines[$i] =~ /^\s*([^:]+):\s*(critical|\s*)$/i) {
			$key = $1;
			$ret->{OPENSSL_EXTENSIONS}->{$key} = [];
            push(@{$ret->{OPENSSL_EXTENSIONS}->{$key}}, $2) if($2 eq "critical");
			$i++;
			while($i < @lines && $lines[$i] !~ /^\s.+:\s*(critical|\s*)$/) {
				$val = $lines[$i];
				$val =~ s/^\s+//g;
				$val =~ s/\s+$//g;
				$i++;
				next if $val =~ /^$/;
				push(@{$ret->{OPENSSL_EXTENSIONS}->{$key}}, $val);
			}
		} else {
			$i++;
		}
	}

	## load special extensions
	$ret->{IS_CA} = 0;
	for my $val (@{$ret->{OPENSSL_EXTENSIONS}->{"X509v3 Basic Constraints"} || []}) {
		if(defined($val) and $val =~ /CA:TRUE/i) {
			$ret->{IS_CA} = 1;
		}
	}

    my $beginCert = "-----BEGIN CERTIFICATE-----";
    my $endCert   = "-----END CERTIFICATE-----";
    ( $ret->{BODY} ) = ( $this->{item} =~ /($beginCert[\S\s\n]+$endCert)/ );
    
    ## if email was not set then we check the subject alternative name
    if (not $ret->{EMAILADDRESS}) {
        my $h = $ret->{OPENSSL_EXTENSIONS}->{"X509v3 Subject Alternative Name"}[0];
        if ($h && $h =~ /^(.*,|)\s*email:/i) {
            ## email are in subjectAltName
            $h =~ s/^(.*,|)\s*email:\s*//ig;
            $h =~ s/\s*$//g;
            $h =~ s/,.*$//g;
            $ret->{EMAILADDRESS} = $h;
        }
    }
    
    if(not defined $ret) {
        return $this->setError("OPENSSL->parseCert: Cannot parse certificate");
    }
	return $ret;
}

sub getParsedCRL {
    my $this = shift;
    my $data = {@_};

    $this->{'crl'}       = $data->{'DATA'};
    $this->{'crlFormat'} = $data->{'INFORM'};
    my $infile = $data->{'INFILE'};

    if (defined $infile && $infile ne "" ) {
        my $tmpLine;
        open( FD, "<$infile" )
          or return $this->setError("OPENSSL->getParsedCRL: Cannot open infile '$infile' for reading.");
        while( $tmpLine = <FD> ) {
            $this->{crl} .= $tmpLine;
        }
        close(FD);
    }

    if( $this->{crl} ne "" ) {
        $this->{item} = $this->{crl};

        my $beginCRL = "-----BEGIN[\\w\\s]+[-]{5}";
        my $endCRL   = "-----END[\\w\\s]+[-]{5}";
        ( $this->{'crl'} ) = ( $this->{'item'} =~ /($beginCRL[\S\s\n]+$endCRL)/ );
        
        my $tmp = $this->parseCRL();

        if ( not $tmp) {
            return $this->setError("OPENSSL->getParsedCRL: Cannot parse CRL ($errmsg)", 
                                   $this->erropenssl());
        } else {
            return $tmp;
        }
    }
    return $this->setError("OPENSSL->getParsedCRL: Cannot parse CRL");
}

sub getTXTCRL {
    my $this = shift;
    my $data = {@_};
    
    my $Data   = $data->{'DATA'}   || undef;
    my $infile = $data->{'INFILE'} || undef;
    my $inform = $data->{'INFORM'} || "PEM";
    
    my %hash = (
                DATATYPE => "CRL",
                INFORM   => $inform,
                OUTFORM  => "TXT"
               );
    if(defined $infile && $infile ne "" && -e $infile) {
        $hash{'INFILE'} = $infile;
    } elsif(defined $Data && $Data ne "") {
        $hash{'DATA'} = $Data;
    } else {
        return $this->setError("No CRL found!");
    }
    
    my $ret = $this->convert( %hash );
    if(not defined $ret) {
        return $this->setError("Cannot convert CRL to TXT-format ".
                               "(".$this->errmsg().")", $this->erropenssl());
    }
    return $ret;
}

sub parseCRL {
    my $this = shift;

    my ( $version, $issuer, $last, $next, $alg, $tmp);
    my @list;
    my @certs;
    
    my ( $head, $body );
    
    my @attList = ( "version", "issuer", "nextUpdate", "lastUpdate", 
                    "signature_algorithm", "revoked", "extensions" );
    
    my $ret = $this->getCRLAttribute(
                                     ATTRIBUTE_LIST => \@attList,
                                     DATA           => $this->{crl},
                                     INFORM         => $this->{crlFormat});
    if (not $ret) {
        return $this->setError("OPENSSL->parseCRL: Cryptobackend fails (".$OPENSSL::errmsg.")");
    }
    
    ## Parse lines ...
    @certs = split ( /\n/i, $ret->{REVOKED} );
    for (my $i=0; $i<scalar @certs; $i++) {
        my $serial = $certs[$i++];
        my $date   = $certs[$i];
        $date =~ s/^\s*(.*)\s*$/$1/;
        my $reason  = "";
        while ($certs[$i+1] =~ /^  /) {
            $certs[++$i] =~ s/^\s*(.*)\s*$/$1/;
            $reason .= $certs[$i]."\n";
        }

        my $entry = {
                     SERIAL => $serial,
                     DATE   => $date };
        $entry->{REASON} = $reason if(defined $reason && $reason ne "");

        @list = ( @list, $entry );
    }
    $ret->{PLAIN_EXTENSIONS} = $ret->{EXTENSIONS};
    delete $ret->{EXTENSIONS};
	my @lines = split(/\n/, $ret->{PLAIN_EXTENSIONS} || "");
    
	my $i = 0;
	while($i < @lines) {
		if($lines[$i] =~ /^\s*([^:]+):\s*(critical|\s*)$/i) {
			my $key = $1;
			$ret->{OPENSSL_EXTENSIONS}->{$key} = [];
            push(@{$ret->{OPENSSL_EXTENSIONS}->{$key}}, $2) if($2 eq "critical");
			$i++;
			while($i < @lines && $lines[$i] !~ /^\s.+:\s*(critical|\s*)$/) {
				my $val = $lines[$i];
				$val =~ s/^\s+//g;
				$val =~ s/\s+$//g;
				$i++;
				next if $val =~ /^$/;
				push(@{$ret->{OPENSSL_EXTENSIONS}->{$key}}, $val);
			}
		} else {
			$i++;
		}
	}

	$i = 0;
	my $acton = "name";
	while ($i < length($ret->{ISSUER})) {
		if (substr($ret->{ISSUER}, $i, 1) =~ /\\/) {
			$i++;
		} elsif(substr($ret->{ISSUER}, $i, 1) =~ /=/) {
			if ($acton eq "value") {
				$ret->{ISSUER} = substr($ret->{ISSUER}, 0, $i)."\\".substr($ret->{ISSUER}, $i);
				$i++;
			} else {
				$acton = "value";
			}
		} elsif(substr ($ret->{ISSUER}, $i, 1) =~ /[,+]/) {
			$acton = "name";
		}
		$i++;
	}
    
	## split ISSUER into ISSUER_HASH
	if ($ret->{ISSUER} =~ /\\/) {
		my $x500_dn = X500::DN->ParseRFC2253($ret->{ISSUER});
		if (not $x500_dn) {
			return $this->setError("OPENSSL->parseCRL: X500::DN failed.");
		}
		my $rdn;
		foreach $rdn ($x500_dn->getRDNs()) {
			next if($rdn->isMultivalued());
			my @attr_types = $rdn->getAttributeTypes();
			my $type  = $attr_types[0];
			my $value = $rdn->getAttributeValue($type);
			push (@{$ret->{ISSUER_HASH}->{uc($type)}}, $value);
		}
	} else {
		my @rdns = split(/,/, $ret->{ISSUER});
		foreach my $rdn (@rdns) {
			my ($type, $value) = split(/=/, $rdn);
			$type =~ s/^\s*//;
			$type =~ s/\s*$//;
			$value =~ s/^\s*//;
			$value =~ s/\s*$//;
			push (@{$ret->{ISSUER_HASH}->{uc($type)}}, $value);
		}
	}


    $ret->{'BODY'}           = $this->{'crl'};
    $ret->{'REVOKED_PARSED'} = \@list ;

    return $ret;
}

#
# pipe_open2send()
#   Starts a command/function in a child process with
#   stdin redirected to a file descriptor (FD).
#
# Parameters:
#   cmd => [ path or function reference, arguments]
#   ret => exit code
#
# Returns: list with child pid and stdin reference or
#   an empty list on pipe creation error or invalid
#   arguments.
#
#   If the real child execution fails (the function
#   returns undef or exec fails), exit code is 1 or
#   the code given in the 'ret' argument.
#
# Note: use full path (absolute or relative) to the
#   command in exec mode (to avoid shell usage)!
#
# Usage:
#   my ($pid, $fd) = pipe_open2send(
#                        cmd => [sub { ... }, @args]
#                    ) or die "can't exec pipe: $!";
# or
#   my ($pid, $fd) = pipe_open2send(
#                        ret => 42,
#                        cmd => ['./cmd', @args], ret => -1
#                    ) or die "can't exec pipe: $!";
#
#   print $fd "...\n";
#   close($fd);
#   waitpid($pid, 0);
#   my $ret = ($? >> 8);
#
sub pipe_open2send(%)
{
    my %args = @_;
    my $ret  = $args{'ret'};
    my @cmd  = @{$args{'cmd'} || []};
    my %env  = %{$args{'env'} || {}};

    unless(defined($ret) and $ret =~ /^[-]?\d+$/) {
        $ret = 1; # has to be a integer
    }
    unless(scalar(@cmd) and defined($cmd[0])
           and ((ref($cmd[0]) eq 'CODE') or
                ($cmd[0] =~ /^[\/|\.\.?\/]\S+$/))) {
        return ();
    }

    my ($fd0, $fd1);
    unless(socketpair($fd0, $fd1, AF_UNIX,
                      SOCK_STREAM, PF_UNSPEC)) {
        return ();
    }
    $fd0->autoflush(1);
    $fd1->autoflush(1);

    my $pid = fork();
    unless(defined($pid)) {
        my $err = $!;
        close($fd0);
        close($fd1);
        $! = $err;
        return ();
    }

    if(0 == $pid) {
        ###
        ### child
        ###

        #
        # redirect stdin to /dev/null
        # (child func my change it)
        # and stdout/stderr to pipes
        #
        close($fd0);
        if(open(STDIN,  '<&'.fileno($fd1))) {
            close($fd1);

            #
            # child with stdin
            # redirected to FD
            #
            if(ref($cmd[0]) eq 'CODE') {
                #
                # start a function
                #
                my $sub = shift(@cmd);
                my $tmp = $sub->(@cmd);
                if(defined($tmp) and
                    $tmp =~ /^[-]?\d+$/) {
                    $ret = $tmp;
                }
            } else {
                #
                # start a command (abs/rel!)
                #
                exec_wrapper(\@cmd, \%env);
            }
        } else {
            close($fd1);
        }
        exit($ret);
    }
    ###
    ### parent
    ###
    close($fd1);
    return ($pid, $fd0);
}

#
# pipe_open2recv2()
#   Starts a command/function in a child process with
#   stdout and stderr redirected to file descriptors.
#
# Parameters:
#   cmd => [ path or function reference, arguments]
#   ret => exit code
#
# Returns: list with child pid, stdout, stderr reference
#   or an empty list on pipe creation error or invalid
#   arguments.
#
#   If the real child execution fails (the function
#   returns undef or exec fails), exit code is 1 or
#   the code given in the 'ret' argument.
#
# Note: use full path (absolute or relative) to the
#   command in exec mode (to avoid shell usage)!
#
# Usage:
#   my ($pid, $out_fd, $err_fd) =
#       pipe_open2recv2(
#           cmd => [sub { ... }, @args]
#       ) or die "can't exec pipe: $!";
# or
#   my ($pid, $out_fd, $err_fd) =
#       pipe_open2recv2(
#           ret => 42,
#           cmd => ['./cmd', @args], ret => -1
#       ) or die "can't exec pipe: $!";
#
#   my ($out, $err);
#   (undef, $out, $err) = pipe_recv2_data(
#                             0, $pid, $out_fd, $err_fd
#                         );
#   waitpid($pid, 0);
#   my $ret = ($? >> 8);
#
sub pipe_open2recv2(%)
{
    my %args = @_;
    my $ret  = $args{'ret'};
    my @cmd  = @{$args{'cmd'} || []};
    my %env  = %{$args{'env'} || {}};

    unless(defined($ret) and $ret =~ /^[-]?\d+$/) {
        $ret = 1; # has to be a integer
    }
    unless(scalar(@cmd) and defined($cmd[0])
           and ((ref($cmd[0]) eq 'CODE') or
                ($cmd[0] =~ /^[\/|\.\.?\/]\S+$/))) {
        return ();
    }

    my ($fd0_out, $fd1_out);
    my ($fd0_err, $fd1_err);
    unless(socketpair($fd0_out, $fd1_out, AF_UNIX,
                      SOCK_STREAM, PF_UNSPEC)) {
        return ();
    }
    unless(socketpair($fd0_err, $fd1_err, AF_UNIX,
                      SOCK_STREAM, PF_UNSPEC)) {
        return ();
    }
    $fd0_err->autoflush(1);
    $fd0_out->autoflush(1);
    $fd1_err->autoflush(1);
    $fd1_out->autoflush(1);

    my $pid = fork();
    unless(defined($pid)) {
        my $err = $!;
        close($fd0_out);
        close($fd0_err);
        close($fd1_out);
        close($fd1_err);
        $! = $err;
        return ();
    }

    if(0 == $pid) {
        ###
        ### child
        ###
        close($fd0_out);
        close($fd0_err);

        #
        # redirect stdin to /dev/null
        # (child func my change it)
        # and stdout/stderr to pipes
        #
        if(open(STDIN,  '<', '/dev/null') and
           open(STDOUT, '>&'.fileno($fd1_out)) and
           open(STDERR, '>&'.fileno($fd1_err)))
        {
            close($fd1_out);
            close($fd1_err);

            if(ref($cmd[0]) eq 'CODE') {
                #
                # start a function
                #
                my $sub = shift(@cmd);
                my $tmp = $sub->(@cmd);
                if(defined($tmp) and
                    $tmp =~ /^[-]?\d+$/) {
                    $ret = $tmp;
                }
            } else {
                #
                # start a command (abs/rel!)
                #
                exec_wrapper(\@cmd, \%env);
            }
        } else {
            close($fd1_out);
            close($fd1_err);
        }
        exit($ret);
    }
    ###
    ### parent
    ###
    close($fd1_out);
    close($fd1_err);
    return ($pid, $fd0_out, $fd0_err);
}

#
# pipe_recv2_data()
#   Reads data from specified file descriptors and
#   returns the data read from them. All fd's are
#   closed before return.
#
# Parameters:
#   first return value argument (pid)
#   list of file descriptors
#
# Returns: a list with its first argument and the
#   data (scalar) buffers received from fd's in
#   the same order as the fd's.
#   On receive error, the data is dropped and the
#   errno restored before return.
#
# Usage:
#   See pipe_open2recv2() description.
#
sub pipe_recv2_data(@)
{
    my $pid = shift;
    my @fds = @_;
    my @out = ();
    my $err = undef;

    if(scalar(@fds)) {
        if(my $sel = IO::Select->new(@fds)) {

            # initialize buffers
            for(my $i=0; $i<scalar(@fds); $i++) {
                $out[$i] = '';
            }

            RUN: while($sel->count()) {

                my @list = $sel->can_read();
                last unless(scalar(@list));

                for(my $i=0; $i<scalar(@fds); $i++) {
                    next unless(grep($_ eq $fds[$i], @list));

                    my $buf = '';
                    my $cnt = $fds[$i]->sysread($buf, 4096);
                    if(defined($cnt)) {
                        if($cnt > 0) {
                            $out[$i] .= $buf;
                        } else {
                            # OK, end of file only
                            $sel->remove($fds[$i]);
                        }
                    } else {
                        # remember i/o error
                        $err = $!;
                        @out = ();
                        last RUN;
                    }
                }
            }
        } else {
            $err = $!; # perhaps there is one
        }

        # close all fd's
        for(my $i=scalar(@fds)-1; $i>=0; $i--) {
            $fds[$i]->close();
        }
        if(defined($err)) {
            # restore errno
            $! = $err;
        }
    }
    return ($pid, @out);
}


#
# write_data()
#   sends data buffer to file handle in blocking mode
#
# Parameters:
#   file handle
#   data buffer
#
# Returns: undef on error or numer of sent bytes
#
sub write_data($$)
{
    my $fout = shift;
    my $data = shift;

    return undef unless($fout and defined($data));
    my $todo = length($data);
    my $done = 0;

    while($todo > $done) {
        my $cnt = syswrite($fout, $data, $todo, $done);
        unless(defined($cnt) and $cnt > 0) {
            return undef;
        }
        $done += $cnt;
    }
    return $done;
}


#
# exec_wrapper()
#   simple exec wrapper to cleanup ENV variables
#   a little bit and (optional) add further ones
#
# Parameters:
#   list reference to a command
#   hash reference to env variables
#
# Returns: does not return, except exec fails,
#          where return code from exec is used
#
sub exec_wrapper(\@\%)
{
    my $cmd = shift;
    my $env = shift;

    $ENV{'PATH'} = '/bin:/usr/bin:/sbin:/usr/sbin';
    delete @ENV{'IFS','CDPATH','ENV','BASH_ENV'};

    # apply some environment vars
    for my $var (keys %$env) {
        $ENV{$var} = $env->{$var};
    }

    exec {@$cmd[0]} @$cmd;
}

=head1 NAME

OPENSSL - a Perl interface to openssl

=head1 SYNOPSIS

  use OPENSSL;

=head1 COPYRIGHT AND LICENSE

Copyright 2004, Novell, Inc.  All rights reserved.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself, either Perl version 5.8.1 or,
at your option, any later version of Perl 5 you may have available.

=cut

1;

